package com.huawei.dli.jdbc;

import com.huawei.dli.jdbc.utils.ColumnInfo;
import com.huawei.dli.jdbc.utils.DescribeTableResponse;
import com.huawei.dli.jdbc.utils.DliColumn;
import com.huawei.dli.jdbc.utils.DliException;
import com.huawei.dli.jdbc.utils.ErrorCode;
import com.huawei.dli.jdbc.utils.GetTablesResponse;
import com.huawei.dli.jdbc.utils.TableStruct;
import com.huawei.dli.jdbc.utils.type.TypeInfo;
import com.huawei.dli.jdbc.utils.type.TypeInfoFactory;
import com.huawei.dli.jdbc.utils.type.TypeInfoParser;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.RowIdLifetime;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;

public class DliDatabaseMetaData implements DatabaseMetaData {

    private static final String PRODUCT_NAME = "Data Lake Insight";

    private static final String PRODUCT_VERSION = "1.1.1";

    private static final String DRIVER_NAME = "DLI";

    private static final int DRIVER_MAJOR_VERSION = 1;

    private static final int DRIVER_MINOR_VERSION = 1;

    private static final int MAX_TABLE_NAME_LENGTH = 128;

    private static final String CATALOG_SEPARATOR = ".";

    private static final char SEARCH_STRING_ESCAPE = '\\';

    private static final String DEFAULT_DLI_CATALOG = "";

    private static final int API_DELAY_PERIOD = 200;

    DliConnection connection;

    public DliDatabaseMetaData(DliConnection connection) {
        this.connection = connection;
    }

    public boolean allProceduresAreCallable() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean allTablesAreSelectable() throws SQLException {
        return true;
    }

    public boolean autoCommitFailureClosesAllResultSets() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean dataDefinitionCausesTransactionCommit() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean dataDefinitionIgnoredInTransactions() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean deletesAreDetected(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean doesMaxRowSizeIncludeBlobs() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getAttributes(String catalog, String schemaPattern,
        String typeNamePattern, String attributeNamePattern) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getBestRowIdentifier(String catalog, String schema,
        String table, int scope, boolean nullable) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public String getCatalogSeparator() throws SQLException {
        return CATALOG_SEPARATOR;
    }

    public String getCatalogTerm() throws SQLException {
        return "instance";
    }

    /**
     * In UQuery, we list all the databases in the service
     *
     * @return ResultSet
     * @throws SQLException If fail to call the API, e.g. server error or other reason.
     */
    public ResultSet getCatalogs() throws SQLException {
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TABLE_CAT"), Arrays.asList(TypeInfoFactory.STRING));
        List<Object[]> rows = new ArrayList<Object[]>();
        return new DliStaticResultSet(connection, meta, rows.iterator());
    }

    public ResultSet getClientInfoProperties() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getColumnPrivileges(String catalog, String schema,
        String table, String columnNamePattern) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getPseudoColumns(String catalog, String schemaPattern,
        String tableNamePattern, String columnNamePattern) throws SQLException {
        // JDK 1.7
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean generatedKeyAlwaysReturned() throws SQLException {
        // JDK 1.7
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    /**
     * Convert a pattern containing JDBC catalog search wildcards into
     * Java regex patterns.
     *
     * @param pattern input which may contain '%' or '_' wildcard characters, or
     *                these characters escaped using {@link #getSearchStringEscape()}.
     * @return replace %/_ with regex search characters, also handle escaped
     * characters.
     */
    private String convertPattern(final String pattern) {
        if (pattern == null) {
            return ".*";
        } else {
            StringBuilder result = new StringBuilder(pattern.length());

            boolean escaped = false;
            for (int i = 0, len = pattern.length(); i < len; i++) {
                char c = pattern.charAt(i);
                if (escaped) {
                    if (c != SEARCH_STRING_ESCAPE) {
                        escaped = false;
                    }
                    result.append(c);
                } else {
                    if (c == SEARCH_STRING_ESCAPE) {
                        escaped = true;
                        continue;
                    } else if (c == '%') {
                        result.append(".*");
                    } else if (c == '_') {
                        result.append('.');
                    } else {
                        result.append(Character.toLowerCase(c));
                    }
                }
            }

            return result.toString();
        }
    }

    public ResultSet getColumns(String catalog, String schemaPattern,
        String tableNamePattern, String columnNamePattern) throws SQLException {

        if (tableNamePattern == null) {
            throw new DliException("Table name must be given when getColumns",
                ErrorCode.JDBC_NO_SUCH_OBJECT_ERROR.toString());
        }

        Statement statement = null;
        List<Object[]> rows = new ArrayList<Object[]>();
        ResultSet tables = null;
        try {
            tables = getTables(catalog, schemaPattern, tableNamePattern, null);
            while (tables.next()) {
                String dbName = tables.getString("TABLE_SCHEM");
                String tableName = tables.getString("TABLE_NAME");

                List<ColumnInfo> columns = null;
                // Sometimes, we may have no permission to describe the specific table,
                // the request will get a DliException
                try {
                    Thread.sleep(DliDatabaseMetaData.API_DELAY_PERIOD);
                    DescribeTableResponse response = connection.getServiceProxy().describeTableInfo(dbName, tableName);
                    if (response != null) {
                        columns = response.getColumns();
                    }
                } catch (DliException e) {
                    continue;
                } catch (InterruptedException e) {
                    throw new SQLException(e.getMessage());
                }
                if (columns != null) {
                    int ordinalPos = 0;
                    for (ColumnInfo info : columns) {
                        ordinalPos++;
                        String columnName = info.getColumnName();
                        if (columnNamePattern != null && !matchPattern(columnName, columnNamePattern)) {
                            continue;
                        }
                        String typeName = info.getType();
                        String remarks = info.getDescription();
                        TypeInfo typeInfo = TypeInfoParser.getTypeInfoFromTypeString(typeName);
                        DliColumn jdbcCol = new DliColumn(columnName, tableName, dbName, typeInfo.getDliType(),
                            typeInfo, remarks, ordinalPos);

                        Object[] rowVals = {
                            null, jdbcCol.getTableSchema(), jdbcCol.getTableName(), jdbcCol.getColumnName(),
                            (long) jdbcCol.getType(), jdbcCol.getTypeName(), null, null,
                            (long) jdbcCol.getDecimalDigits(), (long) jdbcCol.getNumPercRaidx(),
                            (long) jdbcCol.getIsNullable(), jdbcCol.getComment(), null, null, null, null,
                            (long) jdbcCol.getOrdinalPos(), jdbcCol.getIsNullableString(), null, null, null,
                            null, "NO"
                        };
                        rows.add(rowVals);
                    }
                }
            }
        } finally {
            if (tables != null) {
                tables.close();
            }
            if (statement != null) {
                statement.close();
            }
        }

        // Build result set meta data
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TABLE_CAT", "TABLE_SCHEM", "TABLE_NAME",
                "COLUMN_NAME", "DATA_TYPE", "TYPE_NAME", "COLUMN_SIZE", "BUFFER_LENGTH",
                "DECIMAL_DIGITS", "NUM_PERC_RADIX", "NULLABLE", "REMARKS", "COLUMN_DEF",
                "SQL_DATA_TYPE", "SQL_DATETIME_SUB", "CHAR_OCTET_LENGTH", "ORDINAL_POSITION",
                "IS_NULLABLE", "SCOPE_CATALOG", "SCOPE_SCHEMA", "SCOPE_TABLE", "SOURCE_DATA_TYPE", "IS_AUTO_INCREMENT"),
                Arrays.asList(TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                    TypeInfoFactory.STRING, TypeInfoFactory.BIGINT, TypeInfoFactory.STRING,
                    TypeInfoFactory.BIGINT, TypeInfoFactory.BIGINT, TypeInfoFactory.BIGINT,
                    TypeInfoFactory.BIGINT, TypeInfoFactory.BIGINT, TypeInfoFactory.STRING,
                    TypeInfoFactory.STRING, TypeInfoFactory.BIGINT, TypeInfoFactory.BIGINT,
                    TypeInfoFactory.BIGINT, TypeInfoFactory.BIGINT, TypeInfoFactory.STRING,
                    TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                    TypeInfoFactory.BIGINT, TypeInfoFactory.STRING));
        return new DliStaticResultSet(connection, meta, rows.iterator());
    }

    public Connection getConnection() throws SQLException {
        return this.connection;
    }

    public ResultSet getCrossReference(String primaryCatalog,
        String primarySchema, String primaryTable, String foreignCatalog,
        String foreignSchema, String foreignTable) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    public int getDatabaseMajorVersion() throws SQLException {
        return 0;
    }

    public int getDatabaseMinorVersion() throws SQLException {
        return 0;
    }

    public String getDatabaseProductName() throws SQLException {
        return PRODUCT_NAME;
    }

    public String getDatabaseProductVersion() throws SQLException {
        return PRODUCT_VERSION;
    }

    public int getDefaultTransactionIsolation() throws SQLException {
        return Connection.TRANSACTION_NONE;
    }

    public int getDriverMajorVersion() {
        return DRIVER_MAJOR_VERSION;
    }

    public int getDriverMinorVersion() {
        return DRIVER_MINOR_VERSION;
    }

    public String getDriverName() throws SQLException {
        return DRIVER_NAME;
    }

    public String getDriverVersion() throws SQLException {
        return DRIVER_MAJOR_VERSION + "." + DRIVER_MINOR_VERSION;
    }

    public ResultSet getExportedKeys(String catalog, String schema, String table)
        throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public String getExtraNameCharacters() throws SQLException {
        // TODO: verify that this is correct
        return "";
    }

    public ResultSet getFunctionColumns(String arg0, String arg1, String arg2,
        String arg3) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getFunctions(String catalogName, String schemaPattern, String functionNamePattern)
        throws SQLException {
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("FUNCTION_CAT", "FUNCTION_SCHEM", "FUNCTION_NAME",
                "REMARKS", "FUNCTION_TYPE", "SPECIFIC_NAME"), Arrays.asList(TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.BIGINT, TypeInfoFactory.STRING));
        return new DliStaticResultSet(connection, meta);
    }

    public String getIdentifierQuoteString() throws SQLException {
        return " ";
    }

    public ResultSet getImportedKeys(String catalog, String schema, String table)
        throws SQLException {
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("PKTABLE_CAT", "PKTABLE_SCHEM", "PKTABLE_NAME",
                "PKCOLUMN_NAME", "FKTABLE_CAT", "FKTABLE_SCHEM", "FKTABLE_NAME", "FKCOLUMN_NAME",
                "KEY_SEQ", "UPDATE_RULE", "DELETE_RULE", "FK_NAME", "PK_NAME", "DEFERRABILITY"),
                Arrays.asList(TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                    TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                    TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.BIGINT,
                    TypeInfoFactory.BIGINT, TypeInfoFactory.BIGINT, TypeInfoFactory.STRING,
                    TypeInfoFactory.STRING, TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta);
    }

    public ResultSet getIndexInfo(String catalog, String schema, String table,
        boolean unique, boolean approximate) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    public int getJDBCMajorVersion() throws SQLException {
        return 3;
    }

    public int getJDBCMinorVersion() throws SQLException {
        return 0;
    }

    public int getMaxBinaryLiteralLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxCatalogNameLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxCharLiteralLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    /**
     * Returns the value of maxColumnNameLength.
     */
    public int getMaxColumnNameLength() throws SQLException {
        return 128;
    }

    public int getMaxColumnsInGroupBy() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxColumnsInIndex() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxColumnsInOrderBy() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxColumnsInSelect() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxColumnsInTable() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxConnections() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxCursorNameLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxIndexLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxProcedureNameLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxRowSize() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxSchemaNameLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxStatementLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxStatements() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxTableNameLength() throws SQLException {
        return MAX_TABLE_NAME_LENGTH;
    }

    public int getMaxTablesInSelect() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getMaxUserNameLength() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public String getNumericFunctions() throws SQLException {
        return "";
    }

    public ResultSet getPrimaryKeys(String catalog, String schema, String table)
        throws SQLException {
        // Return an empty result set
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TABLE_CAT", "TABLE_SCHEM", "TABLE_NAME",
                "COLUMN_NAME", "KEY_SEQ", "PK_NAME"), Arrays.asList(TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.BIGINT, TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta);
    }

    public ResultSet getProcedureColumns(String catalog, String schemaPattern,
        String procedureNamePattern, String columnNamePattern)
        throws SQLException {
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("STUPID_PLACEHOLDERS", "USELESS_PLACEHOLDER"),
                Arrays.asList(TypeInfoFactory.STRING, TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta);

    }

    public String getProcedureTerm() throws SQLException {
        return "UDF";
    }

    public ResultSet getProcedures(String catalog, String schemaPattern,
        String procedureNamePattern) throws SQLException {
        // Return an empty result set
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("PROCEDURE_CAT", "PROCEDURE_SCHEM",
                "PROCEDURE_NAME", "RESERVERD", "RESERVERD", "RESERVERD", "REMARKS", "PROCEDURE_TYPE",
                "SPECIFIC_NAME"), Arrays.asList(TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.BIGINT,
                TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta);
    }

    public int getResultSetHoldability() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public RowIdLifetime getRowIdLifetime() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public String getSQLKeywords() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public int getSQLStateType() throws SQLException {
        return DatabaseMetaData.sqlStateSQL99;
    }

    public String getSchemaTerm() throws SQLException {
        return "database";
    }

    public ResultSet getSchemas() throws SQLException {
        return getSchemas(null, null);
    }

    /**
     * In UQuery, we list all matched databases as schema
     *
     * @param catalog       the catalog to get schema
     * @param schemaPattern the reg pattern
     * @return ResultSet
     * @throws SQLException If fail to call the API, e.g. server error or other reason.
     */
    public ResultSet getSchemas(String catalog, String schemaPattern)
        throws SQLException {
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TABLE_SCHEM", "TABLE_CATALOG"), Arrays.asList(
                TypeInfoFactory.STRING, TypeInfoFactory.STRING));

        if (schemaPattern == null || schemaPattern.isEmpty()) {
            schemaPattern = "%";
        }

        String currentDb = connection.getConnRes().getDatabaseName();
        Statement statement = null;
        ResultSet rs = null;
        try {
            connection.getConnRes().setDatabaseName(null);
            statement = connection.createStatement();
            rs = statement.executeQuery("show databases;");
            List<Object[]> rows = new ArrayList<Object[]>();
            while (rs.next()) {
                String dbName = (String) rs.getObject(1);
                if (matchPattern(dbName, schemaPattern)) {
                    Object[] obj = {dbName, null};
                    rows.add(obj);
                }
            }
            statement.close();
            return new DliStaticResultSet(connection, meta, rows.iterator());
        } finally {
            if (rs != null) {
                rs.close();
            }
            if (statement != null) {
                statement.close();
            }
            connection.getConnRes().setDatabaseName(currentDb);
        }
    }

    public String getSearchStringEscape() throws SQLException {
        return String.valueOf(SEARCH_STRING_ESCAPE);
    }

    public String getStringFunctions() throws SQLException {
        return "";
    }

    public ResultSet getSuperTables(String catalog, String schemaPattern,
        String tableNamePattern) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getSuperTypes(String catalog, String schemaPattern,
        String typeNamePattern) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public String getSystemFunctions() throws SQLException {
        return "";
    }

    public ResultSet getTablePrivileges(String catalog, String schemaPattern,
        String tableNamePattern) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getTableTypes() throws SQLException {
        List<Object[]> rows = new ArrayList<Object[]>();
        String[] row1 = {"TABLE"};
        String[] row2 = {"VIEW"};
        rows.add(row1);
        rows.add(row2);
        // Build result set meta data
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TABLE_TYPE"), Arrays.asList(TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta, rows.iterator());
    }

    public ResultSet getTables(String catalog, String schemaPattern,
        String tableNamePattern, String[] types) throws SQLException {
        List<Object[]> rows = new ArrayList<Object[]>();
        ResultSet dbs = getSchemas(catalog, schemaPattern);
        try {
            while (dbs.next()) {
                String dbName = dbs.getString("TABLE_SCHEM");
                GetTablesResponse tablesResponse = connection.getServiceProxy().getTablesInfo(dbName);
                for (TableStruct table : tablesResponse.getTables()) {
                    if (matchPattern(table.getTableName(), tableNamePattern)) {
                        String jdbcType = toJdbcTableType(table.getTableType());
                        if (types != null && types.length != 0) {
                            if (!Arrays.asList(types).contains(jdbcType)) {
                                continue;
                            }
                        }
                        Object[] rowVals = {
                                null, dbName, table.getTableName(), jdbcType, table.getTableName(),
                                null, null, null, null,
                                "USER"
                        };
                        rows.add(rowVals);
                    }
                    try {
                        Thread.sleep(DliDatabaseMetaData.API_DELAY_PERIOD);
                    } catch (InterruptedException e) {
                        throw new SQLException(e.getMessage());
                    }
                }
            }
        } finally {
            dbs.close();
        }
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TABLE_CAT", "TABLE_SCHEM", "TABLE_NAME",
                "TABLE_TYPE", "REMARKS", "TYPE_CAT", "TYPE_SCHEM", "TYPE_NAME",
                "SELF_REFERENCING_COL_NAME", "REF_GENERATION"), Arrays.asList(TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta, rows.iterator());
    }

    public static String toJdbcTableType(String dliTabletype) {
        if (dliTabletype == null) {
            return null;
        } else {
            String type = dliTabletype.toUpperCase(Locale.US);
            if (type.equals("MANAGED") || type.equals("EXTERNAL")) {
                return "TABLE";
            } else if (type.equals("VIEW")) {
                return "VIEW";
            } else {
                return type;
            }
        }
    }

    public String getTimeDateFunctions() throws SQLException {
        return "";
    }

    public ResultSet getTypeInfo() throws SQLException {
        // Return an empty result set
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("STUPID_PLACEHOLDERS", "USELESS_PLACEHOLDER"),
                Arrays.asList(TypeInfoFactory.STRING, TypeInfoFactory.STRING));

        return new DliStaticResultSet(connection, meta);
    }

    public ResultSet getUDTs(String catalog, String schemaPattern,
        String typeNamePattern, int[] types) throws SQLException {
        DliResultSetMetaData meta =
            new DliResultSetMetaData(Arrays.asList("TYPE_CAT", "TYPE_SCHEM", "TYPE_NAME",
                "CLASS_NAME", "DATA_TYPE", "REMARKS", "BASE_TYPE"), Arrays.asList(
                TypeInfoFactory.STRING, TypeInfoFactory.STRING, TypeInfoFactory.STRING,
                TypeInfoFactory.STRING, TypeInfoFactory.BIGINT, TypeInfoFactory.STRING,
                TypeInfoFactory.BIGINT));

        return new DliStaticResultSet(connection, meta);
    }

    public String getURL() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public String getUserName() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public ResultSet getVersionColumns(String catalog, String schema, String table)
        throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean insertsAreDetected(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean isCatalogAtStart() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean isReadOnly() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean locatorsUpdateCopy() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean nullPlusNonNullIsNull() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean nullsAreSortedAtEnd() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean nullsAreSortedAtStart() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean nullsAreSortedHigh() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean nullsAreSortedLow() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean othersDeletesAreVisible(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean othersInsertsAreVisible(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean othersUpdatesAreVisible(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean ownDeletesAreVisible(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean ownInsertsAreVisible(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean ownUpdatesAreVisible(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean storesLowerCaseIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean storesLowerCaseQuotedIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean storesMixedCaseIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean storesMixedCaseQuotedIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean storesUpperCaseIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean storesUpperCaseQuotedIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsANSI92EntryLevelSQL() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsANSI92FullSQL() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsANSI92IntermediateSQL() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsAlterTableWithAddColumn() throws SQLException {
        return true;
    }

    public boolean supportsAlterTableWithDropColumn() throws SQLException {
        return false;
    }

    public boolean supportsBatchUpdates() throws SQLException {
        return false;
    }

    public boolean supportsCatalogsInDataManipulation() throws SQLException {
        return false;
    }

    public boolean supportsCatalogsInIndexDefinitions() throws SQLException {
        return false;
    }

    public boolean supportsCatalogsInPrivilegeDefinitions() throws SQLException {
        return false;
    }

    public boolean supportsCatalogsInProcedureCalls() throws SQLException {
        return false;
    }

    public boolean supportsCatalogsInTableDefinitions() throws SQLException {
        return false;
    }

    public boolean supportsColumnAliasing() throws SQLException {
        return true;
    }

    public boolean supportsConvert() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsConvert(int fromType, int toType) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsCoreSQLGrammar() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsCorrelatedSubqueries() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsDataDefinitionAndDataManipulationTransactions()
        throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsDataManipulationTransactionsOnly() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsDifferentTableCorrelationNames() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsExpressionsInOrderBy() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsExtendedSQLGrammar() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsFullOuterJoins() throws SQLException {
        return true;
    }

    public boolean supportsGetGeneratedKeys() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsGroupBy() throws SQLException {
        return true;
    }

    public boolean supportsGroupByBeyondSelect() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsGroupByUnrelated() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsIntegrityEnhancementFacility() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsLikeEscapeClause() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsLimitedOuterJoins() throws SQLException {
        return true;
    }

    public boolean supportsMinimumSQLGrammar() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsMixedCaseIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsMixedCaseQuotedIdentifiers() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsMultipleOpenResults() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsMultipleResultSets() throws SQLException {
        return false;
    }

    public boolean supportsMultipleTransactions() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsNamedParameters() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsNonNullableColumns() throws SQLException {
        return false;
    }

    public boolean supportsOpenCursorsAcrossCommit() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsOpenCursorsAcrossRollback() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsOpenStatementsAcrossCommit() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsOpenStatementsAcrossRollback() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsOrderByUnrelated() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsOuterJoins() throws SQLException {
        return true;
    }

    public boolean supportsPositionedDelete() throws SQLException {
        return false;
    }

    public boolean supportsPositionedUpdate() throws SQLException {
        return false;
    }

    public boolean supportsResultSetConcurrency(int type, int concurrency)
        throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsResultSetHoldability(int holdability)
        throws SQLException {
        return false;
    }

    public boolean supportsResultSetType(int type) throws SQLException {
        return true;
    }

    public boolean supportsSavepoints() throws SQLException {
        return false;
    }

    public boolean supportsSchemasInDataManipulation() throws SQLException {
        return true;
    }

    public boolean supportsSchemasInIndexDefinitions() throws SQLException {
        return false;
    }

    public boolean supportsSchemasInPrivilegeDefinitions() throws SQLException {
        return false;
    }

    public boolean supportsSchemasInProcedureCalls() throws SQLException {
        return false;
    }

    public boolean supportsSchemasInTableDefinitions() throws SQLException {
        return true;
    }

    public boolean supportsSelectForUpdate() throws SQLException {
        return false;
    }

    public boolean supportsStatementPooling() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsStoredFunctionsUsingCallSyntax() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsStoredProcedures() throws SQLException {
        return false;
    }

    public boolean supportsSubqueriesInComparisons() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsSubqueriesInExists() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsSubqueriesInIns() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsSubqueriesInQuantifieds() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsTableCorrelationNames() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsTransactionIsolationLevel(int level)
        throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean supportsTransactions() throws SQLException {
        return false;
    }

    public boolean supportsUnion() throws SQLException {
        return false;
    }

    public boolean supportsUnionAll() throws SQLException {
        return true;
    }

    public boolean updatesAreDetected(int type) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean usesLocalFilePerTable() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean usesLocalFiles() throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public boolean isWrapperFor(Class<?> iface) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    public <T> T unwrap(Class<T> iface) throws SQLException {
        throw new SQLFeatureNotSupportedException("Method not supported");
    }

    private static boolean matchPattern(String s, String pattern) {

        if (pattern == null) {
            return true;
        }

        if (pattern.contains("%") || pattern.contains("_")) {
            // (?<!a)  looks 1 char behind and ensure not equal
            String wildcard = pattern.replaceAll("(?<!\\\\)%", "\\\\w*")
                .replaceAll("(?<!\\\\)_", "\\\\w");

            // escape / and %
            wildcard = wildcard.replace("\\%", "%").replace("\\_", "_");

            if (!s.matches(wildcard)) {
                return false;
            }
        } else {
            if (!s.equals(pattern)) {
                return false;
            }
        }

        return true;
    }

}
