package com.huawei.dli.jdbc;

import com.huawei.dli.jdbc.utils.DliColumn;
import com.huawei.dli.jdbc.utils.type.DliType;
import com.huawei.dli.jdbc.utils.type.TypeInfo;

import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

public class DliResultSetMetaData implements ResultSetMetaData {

    private final List<String> columnNames;

    private final List<DliType> columnTypes;

    private final List<? extends TypeInfo> typeInfos;

    private Map<String, Integer> nameIndexMap;

    private String catalogName = " ";

    private String schemeName = " ";

    private String tableName = " ";

    public DliResultSetMetaData(List<String> columnNames, List<? extends TypeInfo> typeInfos) {
        this.columnNames = columnNames;
        this.typeInfos = typeInfos;
        this.columnTypes = new ArrayList<DliType>();
        if (columnNames != null) {
            for (int i = 0; i < columnNames.size(); i++) {
                columnTypes.add(typeInfos.get(i).getDliType());
            }
        }
    }

    @Override
    public String getCatalogName(int column) throws SQLException {
        return catalogName;
    }

    @Override
    public String getColumnClassName(int column) throws SQLException {
        DliType type = columnTypes.get(toZeroIndex(column));
        return DliColumn.columnClassName(type);
    }

    @Override
    public int getColumnCount() throws SQLException {
        return columnNames.size();
    }

    @Override
    public int getColumnDisplaySize(int column) throws SQLException {
        TypeInfo typeInfo = typeInfos.get(toZeroIndex(column));
        return DliColumn.columnDisplaySize(typeInfo);
    }

    @Override
    public String getColumnLabel(int column) throws SQLException {
        return columnNames.get(toZeroIndex(column));
    }

    @Override
    public String getColumnName(int column) throws SQLException {
        return columnNames.get(toZeroIndex(column));
    }

    @Override
    public int getColumnType(int column) throws SQLException {
        DliType type = columnTypes.get(toZeroIndex(column));
        return DliColumn.dliTypeToSqlType(type);
    }

    @Override
    public String getColumnTypeName(int column) throws SQLException {
        TypeInfo typeInfo = typeInfos.get(toZeroIndex(column));
        return typeInfo.getTypeName();
    }

    @Override
    public int getPrecision(int column) throws SQLException {
        TypeInfo typeInfo = typeInfos.get(toZeroIndex(column));
        return DliColumn.columnPrecision(typeInfo);
    }

    @Override
    public int getScale(int column) throws SQLException {
        DliType type = columnTypes.get(toZeroIndex(column));
        TypeInfo typeInfo = typeInfos.get(toZeroIndex(column));
        return DliColumn.columnScale(type, typeInfo);
    }

    @Override
    public String getSchemaName(int column) throws SQLException {
        return schemeName;
    }

    @Override
    public String getTableName(int column) throws SQLException {
        return tableName;
    }

    @Override
    public boolean isAutoIncrement(int column) throws SQLException {
        return false;
    }

    @Override
    public boolean isCaseSensitive(int column) throws SQLException {
        DliType type = columnTypes.get(toZeroIndex(column));
        return DliColumn.columnCaseSensitive(type);
    }

    @Override
    public boolean isCurrency(int column) throws SQLException {
        return false;
    }

    @Override
    public boolean isDefinitelyWritable(int column) throws SQLException {
        return false;
    }

    // TODO: check
    @Override
    public int isNullable(int column) throws SQLException {
        return ResultSetMetaData.columnNullable;
    }

    @Override
    public boolean isReadOnly(int column) throws SQLException {
        return true;
    }

    @Override
    public boolean isSearchable(int column) throws SQLException {
        return true;
    }

    @Override
    public boolean isSigned(int column) throws SQLException {
        DliType type = columnTypes.get(toZeroIndex(column));
        return DliColumn.columnSigned(type);
    }

    @Override
    public boolean isWritable(int column) throws SQLException {
        return false;
    }

    public void setTableName(String table) {
        tableName = table;
    }

    public void setSchemeName(String scheme) {
        schemeName = scheme;
    }

    public void setCatalogName(String catalog) {
        catalogName = catalog;
    }

    /**
     * Used by other classes for querying the index from column name
     *
     * @param name column name
     * @return column index
     */
    public int getColumnIndex(String name) {
        if (nameIndexMap == null) {
            nameIndexMap = new HashMap<String, Integer>();
            for (int i = 0; i < columnNames.size(); ++i) {
                nameIndexMap.put(columnNames.get(i), i + 1);
                nameIndexMap.put(columnNames.get(i).toLowerCase(Locale.US), i + 1);
            }
        }

        Integer index = nameIndexMap.get(name);
        if (index == null) {
            String lowerName = name.toLowerCase(Locale.US);
            if (lowerName.equals(name)) {
                return -1;
            }

            index = nameIndexMap.get(name);
        }

        if (index == null) {
            return -1;
        }

        return index;
    }

    protected int toZeroIndex(int column) {
        if (column <= 0 || column > columnNames.size()) {
            throw new IllegalArgumentException();
        }
        return column - 1;
    }

    @Override
    public <T> T unwrap(Class<T> iface) throws SQLException {
        return null;
    }

    @Override
    public boolean isWrapperFor(Class<?> iface) throws SQLException {
        return false;
    }

    public TypeInfo getColumnTypeInfo(int columnIndex) {
        return typeInfos.get(toZeroIndex(columnIndex));
    }
}
