package com.huawei.dli.jdbc;

import com.huawei.dli.jdbc.utils.CancelJobResponse;
import com.huawei.dli.jdbc.utils.ConnectionResource;
import com.huawei.dli.jdbc.utils.DliException;
import com.huawei.dli.jdbc.utils.ErrorCode;
import com.huawei.dli.jdbc.utils.ExecuteJobResponse;
import com.huawei.dli.jdbc.utils.QueryExportFetchingSession;
import com.huawei.dli.jdbc.utils.QueryResultFetchingSession;
import com.huawei.dli.jdbc.utils.RequestExecuteSql;
import com.huawei.dli.jdbc.utils.ResultFetchingSession;
import com.huawei.dli.sdk.common.JobResultInfo;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.common.utils.SqlUtils;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.sql.SQLTimeoutException;
import java.sql.SQLWarning;
import java.sql.Statement;
import java.text.Normalizer;
import java.util.Locale;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public class DliStatement implements Statement {

    // ExecuteResult is used to store the execute result
    private class ExecuteResult {
        public boolean hasResultSet;

        public SQLException exception;
    }

    private DliConnection connHandle;

    private ResultSet resultSet = null;

    private boolean isClosed = false;

    private int resultSetMaxRows = Integer.MAX_VALUE;

    private long updateCount = -1;

    private String jobId = null;

    private int queryTimeout = 0;

    // This only for compatible of SQL Workbench tools
    private SQLWarning warning = null;

    public DliStatement(DliConnection conn) throws SQLException {
        this(conn, false);
    }

    DliStatement(DliConnection conn, boolean isResultSetScrollable) throws SQLException {
        this.connHandle = conn;
        // This only for compatible of SQL Workbench tools
        this.warning = new SQLWarning("successful completion");
    }

    @Override
    public ResultSet executeQuery(String sql) throws SQLException {
        if (connHandle.getConnRes().isCheckNoResultQuery()) {
            SqlUtils.checkForNoResult(sql);
        }
        return executeQueryInternal(sql);
    }

    private ResultSet executeQueryInternal(String sql) throws SQLException {
        if (!execute(sql)) {
            throw new SQLException("The query did not generate a result set!");
        }
        return resultSet;
    }

    @Override
    public int executeUpdate(String sql) throws SQLException {
        execute(sql);
        return (int) updateCount;
    }

    @Override
    public void close() throws SQLException {
        if (isClosed) {
            return;
        }
        if (resultSet != null) {
            resultSet.close();
            resultSet = null;
        }
        jobId = null;
        connHandle = null;
        isClosed = true;
    }

    @Override
    public int getMaxFieldSize() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public void setMaxFieldSize(int max) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int getMaxRows() throws SQLException {
        return resultSetMaxRows;
    }

    @Override
    public void setMaxRows(int max) throws SQLException {
        if (max <= 0) {
            throw new SQLException("max must be greater than 0");
        }
        this.resultSetMaxRows = max;
    }

    @Override
    public void setEscapeProcessing(boolean enable) throws SQLException {
        // This method only for compatible of SQL Workbench tools
    }

    @Override
    public int getQueryTimeout() throws SQLException {
        return this.queryTimeout;
    }

    @Override
    public void setQueryTimeout(int seconds) throws SQLException {
        this.queryTimeout = seconds;
    }

    @Override
    public void cancel() throws SQLException {
        connHandle.log.debug("execute cancel");
        if (jobId != null) {
            CancelJobResponse cancelJobResp = null;
            try {
                cancelJobResp = connHandle.getServiceProxy().cancelJob(jobId);
                if (cancelJobResp.isSuccess()) {
                    connHandle.log.warn(String.format("the job : %s is cancelled because of "
                        + "execution timeout.", jobId));
                } else {
                    connHandle.log.warn(
                        String.format("the job : %s is cancel failed, %s.", jobId, cancelJobResp.getMessage()));
                }
            } catch (DliException e) {
                connHandle.log.error(String.format("Cancel failed, %s", e.getMessage()));
            }
        } else {
            connHandle.log.debug("jobId is null");
        }
    }

    @Override
    public SQLWarning getWarnings() throws SQLException {
        // This method only for compatible of SQL Workbench tools
        return warning;
    }

    @Override
    public void clearWarnings() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public void setCursorName(String name) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    class ExecuteTask implements Runnable {

        ExecuteTask(DliStatement statement, String sql) {
            this.statement = statement;
            this.executeResult = new ExecuteResult();
            this.sql = sql;
        }

        DliStatement statement;

        ExecuteResult executeResult;

        String sql;

        @Override
        public void run() {
            try {
                checkClosed();
                beforeExecute();
                try {
                    if (processSetClause(sql)) {
                        executeResult.hasResultSet = false;
                    } else if (processUseClause(sql)) {
                        executeResult.hasResultSet = false;
                    } else {
                        // need to submit job to uquery rest service
                        connHandle.log.debug("submit sql request to DLI rest service.");
                        RequestExecuteSql request = new RequestExecuteSql();
                        request.setConf(connHandle.getSessConfList());
                        request.setCurrentDb(connHandle.getConnRes().getDatabaseName());
                        request.setSql(sql);
                        request.setQueueName(connHandle.getConnRes().getQueueName());
                        ExecuteJobResponse jobResponse = connHandle.getServiceProxy().execute(request);
                        jobId = jobResponse.getJobId();
                        updateCount = 0;

                        if (jobResponse.isSuccess()) {
                            if (SqlUtils.isQuery(sql)) {
                                // If the the job is Sync Mode, get the result set form the Response directly.
                                if (jobResponse.isSyncMode()) {
                                    if (jobResponse.hasResultSet()) {
                                        resultSet = new DliForwardResultSet(
                                            statement, jobResponse.getResultFetchingSession());
                                        executeResult.hasResultSet = true;
                                    }
                                } else {
                                    JobResultInfo jobStatusResp =
                                        waitForJobCompleted(
                                            jobResponse.getJobId(),
                                            connHandle.getConnRes().getJobTimeoutSeconds() * 1000);
                                    int directFetchThreshold =
                                        connHandle.getConnRes().getDirectFetchThreshold();
                                    fetchResult(
                                            jobResponse.getJobId(),
                                            jobStatusResp.getResultCount() <= directFetchThreshold,
                                            jobResponse.getSchema());
                                    executeResult.hasResultSet = true;
                                }
                            } else {
                                JobResultInfo finishedStatus =
                                    waitForJobCompleted(
                                        jobResponse.getJobId(),
                                        connHandle.getConnRes().getJobTimeoutSeconds() * 1000);
                                updateCount = finishedStatus.getResultCount();
                                executeResult.hasResultSet = false;
                            }
                        } else {
                            throw new DliException(jobResponse.getMessage(),
                                ErrorCode.JDBC_DLI_SERVER_ERROR.toString());
                        }
                    }
                } catch (DliException ex) {
                    connHandle.log.error(String.format("Fail to run sql: %s, %s", sql, ex.getMessage()));
                    throw new SQLException(ex);
                } finally {
                    jobId = null;
                }
            } catch (SQLException exception) {
                executeResult.exception = exception;
            }
        }
    }

    @Override
    public boolean execute(final String sql) throws SQLException {
        connHandle.log.debug("Run SQL: " + sql);
        ExecutorService executor = Executors.newSingleThreadExecutor();
        ExecuteTask task = new ExecuteTask(this, sql);
        executor.execute(task);
        executor.shutdown();
        try {
            // if query timeout is not set , we do not limit the execute time
            int timeOut = getQueryTimeout() > 0 ? getQueryTimeout() : Integer.MAX_VALUE;
            if (!executor.awaitTermination(timeOut, TimeUnit.SECONDS)) {
                executor.shutdownNow();
                throw new DliException("The sql request is timeout",
                    ErrorCode.JDBC_DLI_TIMEOUT_ERROR.toString());
            } else {
                if (task.executeResult.exception == null) {
                    return task.executeResult.hasResultSet;
                } else {
                    throw task.executeResult.exception;
                }
            }
        } catch (InterruptedException e) {
            executor.shutdownNow();
            throw new DliException("The sql request is interrupted: " + e.getMessage(),
                ErrorCode.JDBC_DLI_TIMEOUT_ERROR.toString());
        }
    }

    private boolean processSetClause(String sql) {
        sql = Normalizer.normalize(sql, Normalizer.Form.NFKC);
        if (sql.matches("(?i)^(\\s*)(SET)(\\s+)(.*)=(.*);?(\\s*)$")) {
            if (sql.contains(";")) {
                sql = sql.replace(';', ' ');
            }
            int i = sql.toLowerCase(Locale.US).indexOf("set");
            String pairString = sql.substring(i + 3);
            String[] pair = pairString.split("=");
            if (pair.length == 2) {
                String key = pair[0].trim();
                String value = getTrimedString(pair[1]);
                if (key.toLowerCase(Locale.US).startsWith(ConnectionResource.DLI_CONF_PREFIX)) {
                    connHandle.getSessConfMap().put(key, value);
                    connHandle.log.info("set statement property: " + key + "=" + value);
                } else if (key.equals("*") && value.toLowerCase(Locale.US).equals("null")) {
                    connHandle.getSessConfMap().clear();
                    connHandle.log.info("clear all the statement properties");
                } else {
                    connHandle.log.warn(String.format("the property name : '%s' should start with '%s'",
                        key, ConnectionResource.DLI_CONF_PREFIX));
                }
                updateCount = 0;
                return true;
            } else {
                return false;
            }
        }
        return false;
    }

    private boolean processUseClause(String sql) throws SQLException {
        sql = Normalizer.normalize(sql, Normalizer.Form.NFKC);
        if (sql.matches("(?i)^(\\s*)(USE)(\\s+)(.*);?(\\s*)$")) {
            if (sql.contains(";")) {
                sql = sql.replace(';', ' ');
            }
            int i = sql.toLowerCase(Locale.US).indexOf("use");
            String databaseName = getTrimedString(sql.substring(i + 3));
            if (databaseName.length() > 0) {
                if (isDatabaseExist(databaseName)) {
                    connHandle.getConnRes().setDatabaseName(databaseName);
                    connHandle.log.debug("set database to " + databaseName);
                } else {
                    throw new DliException(String.format("database '%s' does not exist", databaseName),
                        ErrorCode.JDBC_NO_SUCH_OBJECT_ERROR.toString());
                }
            } else {
                throw new DliException(String.format("database '%s' name can not be empty", databaseName),
                    ErrorCode.JDBC_NO_SUCH_OBJECT_ERROR.toString());
            }
            updateCount = 0;
            return true;
        }
        return false;
    }

    /**
     * trim empty space and "'"
     *
     * @param str
     * @return
     */
    private String getTrimedString(String str) {
        String trimedStr = str.trim();
        if (trimedStr.startsWith("'")) {
            trimedStr = trimedStr.substring(1);
        }
        if (trimedStr.endsWith("'")) {
            trimedStr = trimedStr.substring(0, trimedStr.lastIndexOf("'"));
        }
        return trimedStr;
    }

    private boolean isDatabaseExist(String databaseName) throws SQLException {
        String currentDb = connHandle.getConnRes().getDatabaseName();
        ResultSet dbList = null;
        try {
            // if the currentDb is not exist, the "show databases" can not get correct result since the JobManager will
            // check if the currentDb is exist. So we do not pass the currentDb to JobManager
            connHandle.getConnRes().setDatabaseName(null);
            dbList = executeQueryInternal("show databases");
            while (dbList.next()) {
                if (databaseName.equalsIgnoreCase(dbList.getString(1))) {
                    return true;
                }
            }
            return false;
        } finally {
            connHandle.getConnRes().setDatabaseName(currentDb);
            if (dbList != null) {
                dbList.close();
            }
        }
    }

    private JobResultInfo waitForJobCompleted(String jobId, long timeoutMillis) throws SQLException {
        JobResultInfo statusResponse;
        try {
            long begin = System.currentTimeMillis();
            int times = 0;
            while (true) {
                String info = String.format("retrying get job status for %s, already tried %d time(s), "
                        + "retry policy is retry up to the job timeout, the job timeout is %d milliseconds", jobId,
                    times++, timeoutMillis);
                connHandle.log.info(info);
                long now = System.currentTimeMillis();
                long tookTimeMillis = now - begin;
                if (tookTimeMillis > timeoutMillis) {
                    String error = "execution time exceeds the maximum execution time, "
                        + "you can increase the job time threshold by setting job timeout.";
                    connHandle.log.error(error);
                    // Cancel the job
                    cancel();
                    throw new SQLTimeoutException(error);
                }
                statusResponse = connHandle.getServiceProxy().getJobStatus(jobId);
                if (statusResponse.getSuccess()) {
                    switch (statusResponse.getJobStatus()) {
                        case FINISHED:
                            connHandle.log.debug("sql status: success");
                            return statusResponse;
                        case FAILED: {
                            String reason = statusResponse.getMessage();
                            connHandle.log.error(String.format("execute job : %s failed: %s", jobId, reason));
                            throw new DliException(reason, ErrorCode.JDBC_DLI_SERVER_ERROR.toString());
                        }
                        case CANCELLED: {
                            String reason = statusResponse.getMessage();
                            connHandle.log.error(String.format("execute job : %s cancelled: %s", jobId, reason));
                            throw new DliException(reason, ErrorCode.JDBC_DLI_SERVER_ERROR.toString());
                        }
                        default:
                            break;
                    }
                    Thread.sleep(connHandle.getConnRes().getMonitorPeriod() * 1000);
                } else {
                    throw new DliException(statusResponse.getMessage(), ErrorCode.JDBC_DLI_SERVER_ERROR.toString());
                }
            }
        } catch (InterruptedException | DliException ex) {
            throw new SQLException(ex.getMessage());
        }
    }

    private void fetchResult(String jobId, boolean useDirectFetch, TableSchema schema) throws DliException,
            SQLException {
        ResultFetchingSession fetchResultSession;
        if (useDirectFetch) {
            fetchResultSession = new QueryResultFetchingSession(jobId, connHandle.getServiceProxy(), schema);
        } else {
            fetchResultSession = new QueryExportFetchingSession(jobId, connHandle.getServiceProxy(), schema);
        }
        resultSet = new DliForwardResultSet(this, fetchResultSession);
    }

    private void fetchResult(String jobId, boolean useDirectFetch) throws DliException,
            SQLException {
        fetchResult(jobId, useDirectFetch, null);
    }

    @Override
    public ResultSet getResultSet() throws SQLException {
        return resultSet;
    }

    @Override
    public synchronized int getUpdateCount() throws SQLException {
        checkClosed();
        return (int) updateCount;
    }

    @Override
    public boolean getMoreResults() throws SQLException {
        return false;
    }

    @Override
    public void setFetchDirection(int direction) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int getFetchDirection() throws SQLException {
        return ResultSet.FETCH_FORWARD;
    }

    @Override
    public void setFetchSize(int rows) throws SQLException {
        connHandle.log.warn("setFetchSize unsupported");
    }

    @Override
    public int getFetchSize() throws SQLException {
        return 0;
    }

    @Override
    public int getResultSetConcurrency() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int getResultSetType() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public void addBatch(String sql) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public void clearBatch() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int[] executeBatch() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public DliConnection getConnection() throws SQLException {
        checkClosed();
        return connHandle;
    }

    @Override
    public boolean getMoreResults(int current) throws SQLException {
        return false;
    }

    @Override
    public ResultSet getGeneratedKeys() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int executeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int executeUpdate(String sql, int[] columnIndexes) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int executeUpdate(String sql, String[] columnNames) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public boolean execute(String sql, int autoGeneratedKeys) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public boolean execute(String sql, int[] columnIndexes) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public boolean execute(String sql, String[] columnNames) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public int getResultSetHoldability() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public boolean isClosed() throws SQLException {
        return isClosed;
    }

    @Override
    public void setPoolable(boolean poolable) throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public boolean isPoolable() throws SQLException {
        return false;
    }

    @Override
    public void closeOnCompletion() throws SQLException {
        throw new SQLFeatureNotSupportedException();
    }

    @Override
    public boolean isCloseOnCompletion() throws SQLException {
        return false;
    }

    @Override
    public <T> T unwrap(Class<T> iface) throws SQLException {
        return null;
    }

    @Override
    public boolean isWrapperFor(Class<?> iface) throws SQLException {
        return false;
    }

    private void beforeExecute() throws SQLException {
        // If the statement re-executes another query, the previously-generated resultSet
        // will be implicit closed. And the corresponding temp table will be dropped as well.
        if (resultSet != null) {
            resultSet.close();
            resultSet = null;
        }
        updateCount = -1;
        isClosed = false;
    }

    protected void checkClosed() throws SQLException {
        if (isClosed) {
            throw new DliException("The statement has been closed",
                ErrorCode.JDBC_CONNECTION_ALREADY_CLOSED_ERROR.toString());
        }
    }
}
