package com.huawei.dli.jdbc.utils;

import com.huawei.dli.sdk.Reader;
import com.huawei.dli.sdk.common.Row;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Use multi reader to read data from OBS to local buffer,
 * in this scenario, the sequence of the result data cannot be guaranteed.
 */
public class ConcurrentIterator implements DownloadIterator {
    private static final Logger logger = LoggerFactory.getLogger(ConcurrentIterator.class);

    private static final int POLL_WAIT_SECONDS = 3;

    private static final int DOTTING_THRESHOLD = 50000;

    private final ArrayBlockingQueue<Row> buffQueue = new ArrayBlockingQueue<>(10000);

    private final List<Reader> readers;

    private final long totalCount;

    private ExecutorService threadPool;

    private long fetchCount = 0;

    public ConcurrentIterator(List<Reader> readers, long totalCount) {
        this.readers = readers;
        if (!readers.isEmpty()) {
            logger.info("there are {} inner readers to fetch data", readers.size());
            this.totalCount = totalCount;
            threadPool = Executors.newFixedThreadPool(readers.size());
            readers.forEach(innerReader -> {
                ProducerTask readerTask = new ProducerTask(buffQueue, innerReader);
                try {
                    threadPool.submit(readerTask);
                } catch (Exception e) {
                    throw new RuntimeException("ConcurrentIterator read task submit error", e);
                }
            });
        } else {
            this.totalCount = 0;
        }
    }

    @Override
    public boolean hasNext() {
        return fetchCount < totalCount;
    }

    @Override
    public Row fetch() throws IOException {
        if (hasNext()) {
            try {
                Row row = buffQueue.poll(POLL_WAIT_SECONDS, TimeUnit.SECONDS);
                if (++fetchCount % DOTTING_THRESHOLD == 0) {
                    logger.info("current buffer size: {}", buffQueue.size());
                }
                return row;
            } catch (InterruptedException e) {
                throw new IOException("Fetch data failed", e);
            }
        }
        throw new RuntimeException("There is no new data to fetch, please check");
    }

    @Override
    public void close() throws IOException {
        for (Reader r : readers) {
            r.close();
        }
        buffQueue.clear();
        if (threadPool != null) {
            threadPool.shutdown();
        }
    }

    @Override
    public long getResultCount() {
        return totalCount;
    }

    static class ProducerTask implements Callable<Void> {
        private static final int OFFER_WAIT_SECONDS = 28; // should be less 30s, or else obs will close the connection

        private final ArrayBlockingQueue<Row> buffer;

        private final Reader reader;

        ProducerTask(ArrayBlockingQueue<Row> buffer, Reader reader) {
            this.buffer = buffer;
            this.reader = reader;
        }

        @Override
        public Void call() throws Exception {
            while (!reader.isClosed() && reader.hasNext()) {
                if (!buffer.offer(reader.read(), OFFER_WAIT_SECONDS, TimeUnit.SECONDS)) {
                    throw new RuntimeException("Write data to buff timeout!!!");
                }
            }
            return null;
        }
    }
}
