package com.huawei.dli.jdbc.utils;

import java.net.URI;
import java.nio.charset.Charset;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.SortedMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import lombok.Getter;
import lombok.Setter;

@Getter
@Setter
public class ConnectionResource {

    public enum AuthenticationMode {
        AKSK,
        TOKEN
    }

    public enum Protocol {
        http,
        https
    }

    private static final Pattern KEY_VALUE_REG = Pattern.compile("([^;]*)=([^;]*)[;]?");

    /**
     * keys to retrieve properties from key-value pairs in urls or prop info.
     */
    public static final String ACCESS_KEY_PROP_KEY = "accesskey";

    public static final String SECRET_KEY_PROP_KEY = "secretkey";

    public static final String REGION_NAME_PROP_KEY = "regionname";

    public static final String SERVICE_NAME_PROP_KEY = "servicename";

    public static final String QUEUE_NAME_PROP_KEY = "queuename";

    public static final String DATABASE_NAME_PROP_KEY = "databasename";

    public static final String MONITOR_PERIOD_PROP_KEY = "monitorperiod";

    public static final String JOB_TIMEOUT_PROP_KEY = "jobtimeout";

    public static final String PROJECT_ID_PROP_KEY = "projectid";

    public static final String LOGCONFFILE_PROP_KEY = "logconffile";

    public static final String CHARSET_PROP_KEY = "charset";

    public static final String TOKEN_PROP_KEY = "token";

    public static final String AUTHENTICATION_MODE_PROP_KEY = "authenticationmode";

    public static final String PROTOCOL_PROP_KEY = "protocol";

    public static final String DLI_CONF_PREFIX = "dli.";

    public static final String USE_PROXY_PROP_KEY = "usehttpproxy";

    public static final String PROXY_HOST_PROP_KEY = "proxyhost";

    public static final String PROXY_PORT_PROP_KEY = "proxyport";

    public static final String SQL_SYNC_MODE_KEY = "dli.sql.syncMode.enabled";

    public static final String CHECK_NO_RESULT_QUERY = "dli.sql.checkNoResultQuery";

    public static final String USE_RETRY_KEY = "useRetry";

    public static final String RETRY_TIMES_KEY = "retryTimes";

    public static final String RETRY_INTERVALS_KEY = "retryIntervals";

    public static final String IAM_ENDPOINT_KEY = "iam.endpoint";

    public static final String OBS_ENDPOINT_KEY = "obs.endpoint";

    public static final String DIRECT_FETCH_THRESHOLD_KEY = "directfetchthreshold";

    public static final String EXPORT_USE_BUFFER_THRESHOLD_KEY = "export.buffered.threshold";

    public static final String EXPORT_INNER_THREAD_NUM_KEY = "export.buffered.threads";

    /**
     * Default value
     */
    private static final String JDBC_LUXOR_URL_PREFIX = "jdbc:luxor://";    // deprecated

    // "jdbc:luxor" will be repreplaced by "jdbc:uquery"
    private static final String JDBC_UQUERY_URL_PREFIX = "jdbc:uquery://";

    private static final String JDBC_DLI_URL_PREFIX = "jdbc:dli://";

    public static final long MONITOR_PERIOD_SECOND = 1;

    public static final long JOB_TIMEOUT_SECOND = 300;

    public static final String SERVICE_VERSION = "v1.0";

    public static final String CHARSET_DEFAULT_VALUE = "UTF-8";

    public static final int RESULT_SET_BUFFER_SIZE = 1000;

    public static final Protocol DEFAULT_PROTOCOL = Protocol.https;

    // AK/SK required
    private String ak;

    private String sk;

    private String regionName;

    private String serviceName;

    // TOKEN Authentication Required
    private String token;

    private String engineType;

    // Http proxy Required
    private boolean useHttpProxy = false;

    private String proxyHost;

    private int proxyPort;

    private Protocol protocol = DEFAULT_PROTOCOL;

    private String host;

    private int port = -1;

    private String queueName;

    private String databaseName;

    private String projectId;

    private String logConfFile;

    private long monitorPeriod = MONITOR_PERIOD_SECOND;

    private long jobTimeoutSeconds = JOB_TIMEOUT_SECOND;

    private String charSet = CHARSET_DEFAULT_VALUE;

    private AuthenticationMode authenticationMode = AuthenticationMode.TOKEN;

    private boolean sqlSyncMode = false;

    private boolean checkNoResultQuery = false;

    private boolean useRetry = false;

    private int retryTimes = 3;

    private int retryIntervals = 30000;

    private String iamEndpoint;

    private String obsEndpoint;

    private int directFetchThreshold = 1000;

    private int useBuffThreshold = 100000;

    private int exportThreadNum = 4;

    public String getIamEndpoint() {
        if (iamEndpoint == null || iamEndpoint.isEmpty()) {
            return String.format("iam.%s.myhuaweicloud.com", regionName);
        }
        return iamEndpoint;
    }

    public String getObsEndpoint() {
        if (obsEndpoint == null || obsEndpoint.isEmpty()) {
            return String.format("obs.%s.myhuaweicloud.com", regionName);
        }
        return obsEndpoint;
    }

    public int getDirectFetchThreshold() {
        return directFetchThreshold;
    }

    public static boolean acceptURL(String url) {
        return (url != null) && (url.startsWith(JDBC_UQUERY_URL_PREFIX) || url.startsWith(JDBC_LUXOR_URL_PREFIX)
            || url.startsWith(JDBC_DLI_URL_PREFIX));
    }

    public static boolean isCharsetName(String charsetName) {
        SortedMap<String, Charset> charsets = Charset.availableCharsets();
        for (Map.Entry<String, Charset> setEntry : charsets.entrySet()) {
            if (setEntry.getKey().equals(charsetName) || setEntry.getValue().aliases().contains(charsetName)) {
                return true;
            }
        }
        return false;
    }

    public String getEndpoint() {
        return port == -1 ? host : host + ":" + port;
    }

    public static ConnectionResource constructConnectionResource(Properties info, String url)
        throws DliException {
        ConnectionResource connectionResource = extractParamsFromUrl(url);
        // load from property info
        if (info != null) {
            connectionResource.loadPropertiesFromConf(info);
        }
        connectionResource.validateParameters();
        return connectionResource;
    }

    /**
     * Parse JDBC connection URL
     * The format of the URL is:
     * jdbc:dli://&lt;endPoint&gt;/projectId?conf_list
     * jdbc:uquery://&lt;endPoint&gt;/projectId?conf_list is derepcated
     * jdbc:luxor://&lt;endPoint&gt;/projectId?conf_list is derepcated
     * where the conf_list is semicolon separated &lt;key&gt;=&lt;val&gt; pairs.
     * the endPoint can be a domain name or &lt;host&gt;:&lt;port&gt;
     * examples
     * jdbc:dli://localhost:8082/testproject?databasename=db1;queuename=default
     * jdbc:dli://uquery.cn-north-1.myhwclouds.com/testproject?databasename=db1;queuename=default
     *
     * @param url the connection url
     * @return ConnectionResource
     * @throws DliException If fail to call the API, e.g. server error or cannot deserialize the response body
     */
    public static ConnectionResource extractParamsFromUrl(String url) throws DliException {
        ConnectionResource connectionResource = new ConnectionResource();

        if (!acceptURL(url)) {
            throw new DliException("Bad URL format: Missing prefix " + JDBC_UQUERY_URL_PREFIX,
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }

        // parse the connection uri
        String uriSchema = "schema://";
        int prefixLength;
        if (url.startsWith(JDBC_LUXOR_URL_PREFIX)) {
            prefixLength = JDBC_LUXOR_URL_PREFIX.length();
        } else if (url.startsWith(JDBC_UQUERY_URL_PREFIX)) {
            prefixLength = JDBC_UQUERY_URL_PREFIX.length();
        } else {
            prefixLength = JDBC_DLI_URL_PREFIX.length();
        }
        String uriFromUrl = url.substring(prefixLength);
        URI jdbcURI;
        try {
            jdbcURI = URI.create(uriSchema + uriFromUrl);
        } catch (IllegalArgumentException ex) {
            throw new DliException(String.format("'%s' should follow the URI 'RFC 2396' standard", uriFromUrl),
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }
        connectionResource.setHost(jdbcURI.getHost());
        connectionResource.setPort(jdbcURI.getPort());

        // get projectid
        String uriPath = jdbcURI.getPath();
        if ((uriPath != null) && !uriPath.isEmpty()) {
            // removing leading '/' returned by getPath()
            uriPath = uriPath.substring(1);
            connectionResource.setProjectId(uriPath);
        }

        // parse key-value pairs
        Map<String, String> parameters = new LinkedHashMap<>();
        String confListStr = jdbcURI.getQuery();
        if (confListStr != null) {
            Matcher confMatcher = KEY_VALUE_REG.matcher(confListStr);
            while (confMatcher.find()) {
                if (parameters.put(confMatcher.group(1), confMatcher.group(2)) != null) {
                    throw new DliException("Bad URL format: Multiple values for property " + confMatcher.group(1),
                        ErrorCode.JDBC_URL_PARSE_ERROR.toString());
                }
            }
        }
        connectionResource.loadPropertiesFromConf(parameters);
        return connectionResource;
    }

    public void loadPropertiesFromConf(Map map) throws DliException {
        if (this.projectId == null && map.get(PROJECT_ID_PROP_KEY) != null) {
            this.setProjectId((String) map.get(PROJECT_ID_PROP_KEY));
        }
        if (map.get(ACCESS_KEY_PROP_KEY) != null) {
            this.setAk((String) map.get(ACCESS_KEY_PROP_KEY));
        }
        if (map.get(SECRET_KEY_PROP_KEY) != null) {
            this.setSk((String) map.get(SECRET_KEY_PROP_KEY));
        }
        if (map.get(REGION_NAME_PROP_KEY) != null) {
            this.setRegionName((String) map.get(REGION_NAME_PROP_KEY));
        }
        if (map.get(SERVICE_NAME_PROP_KEY) != null) {
            this.setServiceName((String) map.get(SERVICE_NAME_PROP_KEY));
        }
        if (map.get(QUEUE_NAME_PROP_KEY) != null) {
            this.setQueueName((String) map.get(QUEUE_NAME_PROP_KEY));
        }
        if (map.get(DATABASE_NAME_PROP_KEY) != null) {
            this.setDatabaseName((String) map.get(DATABASE_NAME_PROP_KEY));
        }
        if (map.get(MONITOR_PERIOD_PROP_KEY) != null) {
            try {
                long period = Long.parseLong((String) map.get(MONITOR_PERIOD_PROP_KEY));
                if (period > 0) {
                    this.setMonitorPeriod(period);
                } else {
                    throw new DliException(
                        String.format("%s can only be an integer greater than 0", MONITOR_PERIOD_PROP_KEY),
                        ErrorCode.JDBC_URL_PARSE_ERROR.toString());
                }
            } catch (NumberFormatException ex) {
                throw new DliException(
                    String.format("%s can only be an integer greater than 0", MONITOR_PERIOD_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(JOB_TIMEOUT_PROP_KEY) != null) {
            try {
                long jobTimeOut = Long.parseLong((String) map.get(JOB_TIMEOUT_PROP_KEY));
                if (jobTimeOut > 0) {
                    this.setJobTimeoutSeconds(jobTimeOut);
                } else {
                    throw new DliException(String.format("%s can only be an integer greater than 0",
                        JOB_TIMEOUT_PROP_KEY), ErrorCode.JDBC_URL_PARSE_ERROR.toString());
                }
            } catch (NumberFormatException ex) {
                throw new DliException(String.format("%s can only be an integer greater than 0", JOB_TIMEOUT_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(LOGCONFFILE_PROP_KEY) != null) {
            this.setLogConfFile((String) map.get(LOGCONFFILE_PROP_KEY));
        }
        if (map.get(CHARSET_PROP_KEY) != null) {
            String charset = (String) map.get(CHARSET_PROP_KEY);
            if (isCharsetName(charset)) {
                this.setCharSet(charset);
            } else {
                throw new DliException(String.format("The charset '%s' is not supported!", charset),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(TOKEN_PROP_KEY) != null) {
            this.setToken((String) map.get(TOKEN_PROP_KEY));
        }
        if (map.get(AUTHENTICATION_MODE_PROP_KEY) != null) {
            try {
                String modeStr = (String) map.get(AUTHENTICATION_MODE_PROP_KEY);
                AuthenticationMode mode = AuthenticationMode.valueOf(modeStr.toUpperCase(Locale.US));
                this.setAuthenticationMode(mode);
            } catch (IllegalArgumentException ex) {
                throw new DliException(String.format("%s only support %s and %s", AUTHENTICATION_MODE_PROP_KEY,
                    AuthenticationMode.TOKEN, AuthenticationMode.AKSK), ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(PROTOCOL_PROP_KEY) != null) {
            try {
                String protocolStr = (String) map.get(PROTOCOL_PROP_KEY);
                Protocol protocolInfo = Protocol.valueOf(protocolStr.toLowerCase(Locale.US));
                this.setProtocol(protocolInfo);
            } catch (IllegalArgumentException ex) {
                throw new DliException(String.format("%s only support %s and %s",
                    PROTOCOL_PROP_KEY, Protocol.http, Protocol.https), ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(USE_PROXY_PROP_KEY) != null) {
            this.setUseHttpProxy(Boolean.parseBoolean((String) map.get(USE_PROXY_PROP_KEY)));
        }
        if (map.get(PROXY_HOST_PROP_KEY) != null) {
            this.setProxyHost((String) map.get(PROXY_HOST_PROP_KEY));
        }
        if (map.get(PROXY_PORT_PROP_KEY) != null) {
            try {
                this.setProxyPort(Integer.parseInt((String) map.get(PROXY_PORT_PROP_KEY)));
            } catch (NumberFormatException ex) {
                throw new DliException(String.format("%s can only be an integer",
                    PROXY_PORT_PROP_KEY), ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (map.get(SQL_SYNC_MODE_KEY) != null) {
            this.setSqlSyncMode(Boolean.parseBoolean((String) map.get(SQL_SYNC_MODE_KEY)));
        }
        if (map.get(CHECK_NO_RESULT_QUERY) != null) {
            this.setCheckNoResultQuery(Boolean.parseBoolean((String) map.get(CHECK_NO_RESULT_QUERY)));
        }

        if (map.get(USE_RETRY_KEY) != null) {
            this.setUseRetry(Boolean.parseBoolean((String) map.get(USE_RETRY_KEY)));
        }
        if (map.get(RETRY_TIMES_KEY) != null) {
            this.setRetryTimes(Integer.parseInt((String) map.get(RETRY_TIMES_KEY)));
        }
        if (map.get(RETRY_INTERVALS_KEY) != null) {
            this.setRetryIntervals(Integer.parseInt((String) map.get(RETRY_INTERVALS_KEY)));
        }
        if (map.get(IAM_ENDPOINT_KEY) != null) {
            this.iamEndpoint = (String) map.get(IAM_ENDPOINT_KEY);
        }
        if (map.get(OBS_ENDPOINT_KEY) != null) {
            this.obsEndpoint = (String) map.get(OBS_ENDPOINT_KEY);
        }
        if (map.get(DIRECT_FETCH_THRESHOLD_KEY) != null) {
            this.directFetchThreshold = Integer.parseInt((String) map.get(DIRECT_FETCH_THRESHOLD_KEY));
        }
        if (map.get(EXPORT_USE_BUFFER_THRESHOLD_KEY) != null) {
            this.useBuffThreshold = Integer.parseInt((String) map.get(EXPORT_USE_BUFFER_THRESHOLD_KEY));
        }
        if (map.get(EXPORT_INNER_THREAD_NUM_KEY) != null) {
            this.exportThreadNum = Integer.parseInt((String) map.get(EXPORT_INNER_THREAD_NUM_KEY));
        }
    }

    private void validateParameters() throws DliException {
        String errorMsg;
        if (authenticationMode == AuthenticationMode.TOKEN) {
            if (token == null || token.isEmpty()) {
                errorMsg = String.format("The configuration item '%s' must be set in TOKEN authentication mode",
                    TOKEN_PROP_KEY);
                throw new DliException(errorMsg, ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        } else if (authenticationMode == AuthenticationMode.AKSK) {
            errorMsg = "The configuration item '%s' must be set in AK/SK authentication mode";
            if (ak == null || ak.isEmpty()) {
                throw new DliException(String.format(errorMsg, ACCESS_KEY_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
            if (sk == null || sk.isEmpty()) {
                throw new DliException(String.format(errorMsg, SECRET_KEY_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
            if (regionName == null || regionName.isEmpty()) {
                throw new DliException(String.format(errorMsg, REGION_NAME_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }
        if (projectId == null || projectId.isEmpty()) {
            throw new DliException(String.format("%s must be set", PROJECT_ID_PROP_KEY),
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }
        if (queueName == null || queueName.isEmpty()) {
            throw new DliException(String.format("%s name must be set", QUEUE_NAME_PROP_KEY),
                ErrorCode.JDBC_URL_PARSE_ERROR.toString());
        }
        if (this.useHttpProxy) {
            if (this.proxyHost == null) {
                throw new DliException(String.format("%s must be set when use http proxy", PROXY_HOST_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
            if (this.proxyPort < 0) {
                throw new DliException(String.format("%s must be set greater than 0", PROXY_PORT_PROP_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        }

        if (this.useRetry) {
            if (this.retryTimes < 1) {
                throw new DliException(String.format("%s must be set greater than 1", RETRY_TIMES_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
            if (this.retryIntervals <= 0) {
                throw new DliException(String.format("%s must be set greater than 0", RETRY_INTERVALS_KEY),
                    ErrorCode.JDBC_URL_PARSE_ERROR.toString());
            }
        } else {
            retryTimes = 1;
        }
    }
}
