package com.huawei.dli.jdbc.utils;

import com.huawei.dli.jdbc.utils.type.DecimalTypeInfo;
import com.huawei.dli.jdbc.utils.type.DliType;
import com.huawei.dli.jdbc.utils.type.DliTypeTransformer;
import com.huawei.dli.jdbc.utils.type.TypeInfo;

import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.HashMap;
import java.util.Map;

public class DliColumn {

    public static final String DLI_DATETIME_FORMAT = "yyyy-MM-dd HH:mm:ss";

    public static final int DLI_DECIMAL_PRECISON = 54;

    public static final int DLI_DECIMAL_SCALE = 18;

    public static final int DLI_STRING_CHARACTERS = 8 * 1024 * 1024 / 3;

    private static Map<DliType, Integer> dliTypeIntegerHashMap = new HashMap<DliType, Integer>();

    static {
        dliTypeIntegerHashMap.put(DliType.BIGINT, java.sql.Types.BIGINT);
        dliTypeIntegerHashMap.put(DliType.STRING, java.sql.Types.VARCHAR);
        dliTypeIntegerHashMap.put(DliType.DOUBLE, java.sql.Types.DOUBLE);
        dliTypeIntegerHashMap.put(DliType.DECIMAL, java.sql.Types.DECIMAL);
        dliTypeIntegerHashMap.put(DliType.INT, java.sql.Types.INTEGER);
        dliTypeIntegerHashMap.put(DliType.SMALLINT, java.sql.Types.SMALLINT);
        dliTypeIntegerHashMap.put(DliType.DATE, java.sql.Types.DATE);
        dliTypeIntegerHashMap.put(DliType.TIMESTAMP, java.sql.Types.TIMESTAMP);
        dliTypeIntegerHashMap.put(DliType.CHAR, java.sql.Types.CHAR);
        dliTypeIntegerHashMap.put(DliType.VARCHAR, java.sql.Types.VARCHAR);
        dliTypeIntegerHashMap.put(DliType.SHORT, java.sql.Types.SMALLINT);
        dliTypeIntegerHashMap.put(DliType.LONG, java.sql.Types.BIGINT);
        dliTypeIntegerHashMap.put(DliType.BOOLEAN, java.sql.Types.BOOLEAN);
        dliTypeIntegerHashMap.put(DliType.FLOAT, java.sql.Types.FLOAT);
        dliTypeIntegerHashMap.put(DliType.TINYINT, java.sql.Types.TINYINT);
        dliTypeIntegerHashMap.put(DliType.ARRAY, java.sql.Types.ARRAY);
        dliTypeIntegerHashMap.put(DliType.STRUCT, java.sql.Types.STRUCT);
        dliTypeIntegerHashMap.put(DliType.MAP, java.sql.Types.VARCHAR);
        dliTypeIntegerHashMap.put(DliType.NULL, java.sql.Types.NULL);
    }

    private final String columnName;

    private final String tableName;

    private final String tableSchema;

    private final DliType type;

    private final String comment;

    private final int ordinalPos;

    private final TypeInfo typeInfo;

    public DliColumn(String columnName, String tableName, String tableSchema, DliType type,
        TypeInfo typeInfo, String comment, int ordinalPos) {
        this.columnName = columnName;
        this.tableName = tableName;
        this.tableSchema = tableSchema;
        this.type = type;
        this.typeInfo = typeInfo;
        this.comment = comment;
        this.ordinalPos = ordinalPos;
    }

    public static int dliTypeToSqlType(DliType type) throws SQLException {
        if (dliTypeIntegerHashMap.containsKey(type)) {
            return dliTypeIntegerHashMap.get(type);
        }
        throw new SQLException("Invalid DLI type: " + type);
    }

    public static String columnClassName(DliType type) throws SQLException {
        return DliTypeTransformer.dliTypeToJavaType(type).getName();
    }

    public static int columnDisplaySize(TypeInfo typeInfo) throws SQLException {
        // according to dliTypeToSqlType possible options are:
        int columnType = dliTypeToSqlType(typeInfo.getDliType());
        switch (columnType) {
            case Types.NULL:
                return 4;
            case Types.CHAR:
            case Types.VARCHAR:
            case Types.BOOLEAN:
                return columnPrecision(typeInfo);
            case Types.TINYINT:
            case Types.SMALLINT:
            case Types.INTEGER:
            case Types.BIGINT:
                return columnPrecision(typeInfo) + 1; // allow +/-
            case Types.DATE:
                return 10;
            case Types.TIMESTAMP:
                return columnPrecision(typeInfo);
            // see
            // http://download.oracle.com/javase/6/docs/api/constant-values.html#java.lang.Float.MAX_EXPONENT
            case Types.FLOAT:
                return 24; // e.g. -(17#).e-###
            // see
            // http://download.oracle.com/javase/6/docs/api/constant-values.html#java.lang.Double.MAX_EXPONENT
            case Types.DOUBLE:
                return 25; // e.g. -(17#).e-####
            case Types.DECIMAL:
                return columnPrecision(typeInfo) + 2; // '-' sign and '.'
            case Types.OTHER:
            case Types.JAVA_OBJECT:
                return columnPrecision(typeInfo);
            case Types.ARRAY:
            case Types.STRUCT:
                return Integer.MAX_VALUE;
            default:
                throw new SQLException("Invalid DLI type: " + columnType);
        }
    }

    public static int columnPrecision(TypeInfo typeInfo) throws SQLException {
        int columnType = dliTypeToSqlType(typeInfo.getDliType());
        // according to dliTypeToSqlType possible options are:
        switch (columnType) {
            case Types.NULL:
                return 0;
            case Types.BOOLEAN:
                return 1;
            case Types.CHAR:
            case Types.VARCHAR:
                return Integer.MAX_VALUE;
            case Types.BINARY:
                return Integer.MAX_VALUE;
            case Types.TINYINT:
                return 3;
            case Types.SMALLINT:
                return 5;
            case Types.INTEGER:
                return 10;
            case Types.BIGINT:
                return 19;
            case Types.FLOAT:
                return 7;
            case Types.DOUBLE:
                return 15;
            case Types.DATE:
                return 10;
            case Types.TIMESTAMP:
                return 29;
            case Types.DECIMAL: {
                if (typeInfo instanceof DecimalTypeInfo) {
                    return ((DecimalTypeInfo) typeInfo).getPrecision();
                } else {
                    throw new SQLException("Invalid DLI type: " + columnType);
                }
            }
            case Types.ARRAY:
                return Integer.MAX_VALUE;
            default:
                throw new SQLException("Invalid DLI type: " + columnType);
        }
    }

    public static int columnScale(DliType type, TypeInfo typeInfo) throws SQLException {
        int columnType = dliTypeToSqlType(type);
        // according to dliTypeToSqlType possible options are:
        switch (columnType) {
            case Types.NULL:
            case Types.BOOLEAN:
            case Types.CHAR:
            case Types.VARCHAR:
            case Types.TINYINT:
            case Types.SMALLINT:
            case Types.INTEGER:
            case Types.BIGINT:
            case Types.DATE:
            case Types.BINARY:
                return 0;
            case Types.FLOAT:
                return 7;
            case Types.DOUBLE:
                return 15;
            case Types.TIMESTAMP:
                return 9;
            case Types.DECIMAL: {
                if (typeInfo instanceof DecimalTypeInfo) {
                    return ((DecimalTypeInfo) typeInfo).getScale();
                } else {
                    throw new SQLException("Invalid DLI type: " + columnType);
                }
            }
            case Types.ARRAY:
                return 0;
            default:
                throw new SQLException("Invalid DLI type: " + columnType);
        }
    }

    public static boolean columnCaseSensitive(DliType type) throws SQLException {
        int columnType = dliTypeToSqlType(type);
        // according to dliTypeToSqlType possible options are:
        switch (columnType) {
            case Types.NULL:
            case Types.BOOLEAN:
            case Types.TINYINT:
            case Types.SMALLINT:
            case Types.INTEGER:
            case Types.BIGINT:
            case Types.DATE:
            case Types.BINARY:
            case Types.FLOAT:
            case Types.DOUBLE:
            case Types.TIMESTAMP:
            case Types.DECIMAL:
            case Types.OTHER:
            case Types.JAVA_OBJECT:
            case Types.ARRAY:
            case Types.STRUCT:
                return false;
            case Types.CHAR:
            case Types.VARCHAR:
                return true;
            default:
                throw new SQLException("Invalid DLI type: " + columnType);
        }
    }

    public static boolean columnSigned(DliType type) throws SQLException {
        int columnType = dliTypeToSqlType(type);
        // according to dliTypeToSqlType possible options are:
        switch (columnType) {
            case Types.NULL:
            case Types.BOOLEAN:
            case Types.DATE:
            case Types.BINARY:
            case Types.TIMESTAMP:
            case Types.ARRAY:
            case Types.STRUCT:
            case Types.CHAR:
            case Types.VARCHAR:
                return false;
            case Types.TINYINT:
            case Types.SMALLINT:
            case Types.INTEGER:
            case Types.BIGINT:
            case Types.FLOAT:
            case Types.DOUBLE:
            case Types.DECIMAL:
                return true;
            default:
                throw new SQLException("unknown DliType");
        }
    }

    public String getColumnName() {
        return columnName;
    }

    public String getTableName() {
        return tableName;
    }

    public String getTableSchema() {
        return tableSchema;
    }

    public int getType() throws SQLException {
        return dliTypeToSqlType(type);
    }

    public String getComment() {
        return comment;
    }

    public String getTypeName() {
        return typeInfo.getTypeName();
    }

    public int getOrdinalPos() {
        return ordinalPos;
    }

    public int getNumPercRaidx() {
        return 10;
    }

    public int getDecimalDigits() {
        return 0;
    }

    public int getIsNullable() {
        return DatabaseMetaData.columnNullable;
    }

    public String getIsNullableString() {
        switch (getIsNullable()) {
            case (DatabaseMetaData.columnNoNulls):
                return "NO";
            case (DatabaseMetaData.columnNullable):
                return "YES";
            case (DatabaseMetaData.columnNullableUnknown):
                return null;
            default:
                return null;
        }
    }
}
