package com.huawei.dli.jdbc.utils;

import com.huawei.dli.sdk.DLIClient;
import com.huawei.dli.sdk.Database;
import com.huawei.dli.sdk.DownloadJob;
import com.huawei.dli.sdk.Queue;
import com.huawei.dli.sdk.SQLJob;
import com.huawei.dli.sdk.Table;
import com.huawei.dli.sdk.authentication.AuthenticationMode;
import com.huawei.dli.sdk.common.DLIInfo;
import com.huawei.dli.sdk.common.JobResultInfo;
import com.huawei.dli.sdk.common.Row;
import com.huawei.dli.sdk.common.TableSchema;
import com.huawei.dli.sdk.common.utils.SqlUtils;
import com.huawei.dli.sdk.exception.DLIException;

import java.util.List;

public class DliServiceProxy implements IDliServiceProxy {
    private final String projectId;

    private final String serviceVersion;

    private final String queueName;

    private final String charSet;

    private final int jobTimeout;

    private final int bufferedThreshold;

    private final int bufferedThreads;

    private final DLIClient dliClient;

    public DliServiceProxy(ConnectionResource connRes) {
        this.projectId = connRes.getProjectId();
        this.serviceVersion = ConnectionResource.SERVICE_VERSION;
        this.queueName = connRes.getQueueName();
        this.charSet = connRes.getCharSet();
        this.jobTimeout = (int) connRes.getJobTimeoutSeconds();
        this.bufferedThreshold = connRes.getUseBuffThreshold();
        this.bufferedThreads = connRes.getExportThreadNum();

        DLIInfo dliInfo = new DLIInfo();
        dliInfo.setRegionName(connRes.getRegionName());
        dliInfo.setProjectId(connRes.getProjectId());
        dliInfo.setIamEndPoint(connRes.getIamEndpoint());
        dliInfo.setObsEndPoint(connRes.getObsEndpoint());
        dliInfo.setServerEndPoint(connRes.getEndpoint());
        if (connRes.isUseHttpProxy()) {
            dliInfo.setProxy(connRes.getProxyHost(), connRes.getProxyPort());
        }
        dliInfo.setUseRetry(true);
        if (connRes.getAuthenticationMode() == ConnectionResource.AuthenticationMode.TOKEN) {
            dliInfo.setToken(connRes.getToken());
            dliClient = new DLIClient(AuthenticationMode.TOKEN, dliInfo);
        } else {
            dliInfo.setAccessKey(connRes.getAk());
            dliInfo.setSecretKey(connRes.getSk());
            dliClient = new DLIClient(AuthenticationMode.AKSK, dliInfo);
        }
    }

    public String getProjectId() {
        return projectId;
    }

    public String getServiceVersion() {
        return serviceVersion;
    }

    public String getQueueName() {
        return queueName;
    }

    public String getCharSet() {
        return charSet;
    }

    @Override
    public ExecuteJobResponse execute(RequestExecuteSql request) throws DliException {
        try {
            Queue queue = new Queue(queueName, dliClient);
            SQLJob sqlJob = new SQLJob(queue, request.getCurrentDb(), request.getSql());
            sqlJob.setJobTimeout(jobTimeout);
            sqlJob.submit();
            TableSchema resultSchema = null;
            List<Row> jobResult = null;
            if (SqlUtils.isQuery(sqlJob.getSql())) {
                resultSchema = sqlJob.getResultSchema();
                jobResult = sqlJob.getJobResult();
            }
            return new ExecuteJobResponse(sqlJob.getJobId(), sqlJob.getJobType(), sqlJob.getJobMode(),
                resultSchema, jobResult);
        } catch (DLIException e) {
            return new ExecuteJobResponse(false, e.getMessage());
        }
    }

    @Override
    public JobResultInfo getJobStatus(String jobId) throws DliException {
        try {
            return dliClient.queryJobResultInfo(jobId);
        } catch (DLIException e) {
            throw new DliException(e);
        }
    }

    @Override
    public TableSchema getResultSchema(String jobId) throws DliException {
        try {
            Queue queue = new Queue(queueName, dliClient);
            SQLJob sqlJob = new SQLJob(queue, "");
            sqlJob.setJobId(jobId);
            return sqlJob.getResultSchema();
        } catch (DLIException e) {
            throw new DliException(e);
        }
    }

    @Override
    public GetJobResultResponse getJobResult(String jobId, int pageSize, int currentPage) throws DliException {
        try {
            Queue queue = new Queue(queueName, dliClient);
            SQLJob sqlJob = new SQLJob(queue, "");
            sqlJob.setJobId(jobId);
            GetJobResultResponse getJobResultResp = new GetJobResultResponse();
            getJobResultResp.setRows(sqlJob.getJobResult(pageSize, currentPage));
            getJobResultResp.setRowCount(sqlJob.getRowCount());
            return getJobResultResp;
        } catch (DLIException e) {
            throw new DliException(e);
        }
    }

    @Override
    public DownloadIterator exportJobResult(String jobId) throws DliException {
        try {
            Queue queue = new Queue(queueName, dliClient);
            SQLJob sqlJob = new SQLJob(queue, "");
            sqlJob.setJobId(jobId);
            DownloadJob downloadJob = new DownloadJob(queue, sqlJob);
            com.huawei.dli.sdk.common.JobStatus status =
                com.huawei.dli.sdk.common.JobStatus.fromName(downloadJob.getDownloadStatus());
            while (!(status == com.huawei.dli.sdk.common.JobStatus.FINISHED
                || status == com.huawei.dli.sdk.common.JobStatus.FAILED
                || status == com.huawei.dli.sdk.common.JobStatus.CANCELLED)) {
                try {
                    Thread.sleep(1000L);
                } catch (InterruptedException e) {
                    // ignore
                }
                status = com.huawei.dli.sdk.common.JobStatus.fromName(downloadJob.getDownloadStatus());
            }
            long resultCount = downloadJob.getResultCount();
            if (resultCount > bufferedThreshold) {
                return new ConcurrentIterator(downloadJob.createReaders(bufferedThreads), resultCount);
            }
            return new CommonIterator(downloadJob.createReader(), resultCount);
        } catch (DLIException e) {
            throw new DliException(e);
        }
    }

    @Override
    public CancelJobResponse cancelJob(String jobId) throws DliException {
        try {
            dliClient.cancelJob(jobId);
            return new CancelJobResponse(true, "");
        } catch (DLIException e) {
            return new CancelJobResponse(false, e.getMessage());
        }
    }

    @Override
    public GetTablesResponse getTablesInfo(String dbName) throws DliException {
        try {
            Database db = dliClient.getDatabase(dbName);
            List<Table> tblList = db.listAllTables();
            return new GetTablesResponse(tblList);
        } catch (DLIException e) {
            return new GetTablesResponse(false, e.getMessage());
        }
    }

    @Override
    public DescribeTableResponse describeTableInfo(String dbName, String tblName) throws DliException {
        try {
            Database db = dliClient.getDatabase(dbName);
            Table tbl = db.getTable(tblName);
            return new DescribeTableResponse(tbl.getTableDetail().getColumn());
        } catch (DLIException e) {
            return new DescribeTableResponse(false, e.getMessage());
        }
    }
}
