package com.huawei.dli.jdbc.utils.type;

import java.util.ArrayList;
import java.util.Locale;

public class TypeInfoParser {
    private ArrayList tokens = new ArrayList();

    private String typeInfoName;

    private int index = 0;

    TypeInfoParser(String name) {
        this.typeInfoName = name.toUpperCase(Locale.US);
        this.tokenize(this.typeInfoName);
    }

    private boolean isTypeInfoChar(char c) {
        return Character.isLetterOrDigit(c) || c == '_' || c == '.';
    }

    @SuppressWarnings("unchecked")
    private void tokenize(String name) {
        if (name.startsWith("\"")) {
            name = name.substring(1);
        }
        if (name.endsWith("\"")) {
            name = name.substring(0, name.length() - 1);
        }
        int begin = 0;

        for (int end = 1; end <= name.length(); ++end) {
            if (end == name.length() || !this.isTypeInfoChar(name.charAt(end))
                || !this.isTypeInfoChar(name.charAt(end - 1))) {
                String token = name.substring(begin, end).trim();
                if (!token.isEmpty()) {
                    this.tokens.add(token);
                }

                begin = end;
            }
        }
    }

    private TypeInfo parseTypeInfo() {
        TypeInfo type = this.parseTypeInfoInternal();
        if (this.index != this.tokens.size()) {
            throw new IllegalArgumentException("Parse type info failed, pls check: " + this.typeInfoName);
        } else {
            return type;
        }
    }

    private TypeInfo parseTypeInfoInternal() {
        DliType typeCategory = null;
        try {
            typeCategory = DliType.valueOf(this.peek());
        } catch (IllegalArgumentException e) {
            typeCategory = DliType.NULL;
        }
        switch (typeCategory) {
            case DECIMAL:
                return this.parseComplexDecimalTypeInfo();
            case STRUCT:
                return this.parseStructTypeInfo();
            case ARRAY:
                return this.parseArrayTypeInfo();
            case MAP:
                return this.parseMapTypeInfo();
            default:
                return TypeInfoFactory.getPrimitiveTypeInfo(typeCategory);
        }
    }

    private TypeInfo parseArrayTypeInfo() {
        expect("<");
        TypeInfo elementTypeInfo = parseTypeInfoInternal();
        expect(">");
        return TypeInfoFactory.getArrayTypeInfo(elementTypeInfo);
    }

    private TypeInfo parseMapTypeInfo() {
        expect("<");
        TypeInfo keyTypeInfo = parseTypeInfoInternal();
        expect(",");
        TypeInfo valueTypeInfo = parseTypeInfoInternal();
        expect(">");
        return TypeInfoFactory.getMapTypeInfo(keyTypeInfo, valueTypeInfo);
    }

    private String peek() {
        if (this.index >= this.tokens.size()) {
            throw new IllegalArgumentException("Parse type info failed, pls check: " + this.typeInfoName);
        } else {
            return (String) this.tokens.get(this.index++);
        }
    }

    private int getPosition(int tokenIndex) {
        int pos = 0;

        for (int i = 0; i < tokenIndex; ++i) {
            pos += ((String) this.tokens.get(i)).length();
        }

        return pos;
    }

    private void expect(String expected) {
        if (!this.peek().equals(expected)) {
            throw new IllegalArgumentException("Error parse type info: " + this.typeInfoName
                + ", expect '" + expected + "' in position: " + this.getPosition(this.index - 1));
        }
    }

    private int getInteger() {
        try {
            return Integer.parseInt(this.peek());
        } catch (NumberFormatException var2) {
            throw new IllegalArgumentException("Error parse type info: " + this.typeInfoName
                + ", expect integer in position: " + this.getPosition(this.index - 1), var2);
        }
    }

    private int getCharParam() {
        this.expect("(");
        int length = this.getInteger();
        this.expect(")");
        return length;
    }

    private int[] getDecimalParams() {
        if (this.index < this.tokens.size() && ((String) this.tokens.get(this.index)).equals("(")) {
            this.expect("(");
            int precision = this.getInteger();
            this.expect(",");
            int scale = this.getInteger();
            this.expect(")");
            return new int[] {precision, scale};
        } else {
            return new int[0];
        }
    }

    private TypeInfo parseVarcharTypeInfo() {
        return TypeInfoFactory.getVarcharTypeInfo(this.getCharParam());
    }

    private TypeInfo parseCharTypeInfo() {
        return TypeInfoFactory.getCharTypeInfo(this.getCharParam());
    }

    private TypeInfo parseComplexDecimalTypeInfo() {
        int[] params = this.getDecimalParams();
        return params.length == 0 ? TypeInfoFactory.DECIMAL : TypeInfoFactory.getDecimalTypeInfo(params[0], params[1]);
    }

    private TypeInfo parseStructTypeInfo() {
        ArrayList<String> names = new ArrayList<String>();
        ArrayList<TypeInfo> typeInfos = new ArrayList<TypeInfo>();
        boolean first = true;
        while (true) {
            if (first) {
                expect("<");
                first = false;
            } else {
                String delimiter = peek();
                if (delimiter.equals(">")) {
                    break;
                } else {
                    index--;
                    expect(",");
                }
            }

            names.add(peek());
            expect(":");
            typeInfos.add(parseTypeInfoInternal());
        }
        return TypeInfoFactory.getStructTypeInfo(names, typeInfos);
    }

    public static TypeInfo getTypeInfoFromTypeString(String name) {
        return (new TypeInfoParser(name)).parseTypeInfo();
    }
}
